#include "stdafx.h"
#include "cphidgetadvancedservo.h"
#include "cusb.h"
#include "csocket.h"
#include "cthread.h"

// === Internal Functions === //

//initAfterOpen - sets up the initial state of an object, reading in packets from the device if needed
//				  used during attach initialization - on every attach
CPHIDGETINIT(AdvancedServo)
	int i = 0;
	int readtries = 0;

	TESTPTR(phid);

	//Setup max/min values
	switch(phid->phid.deviceIDSpec)
	{
		case PHIDSPEC_ADVANCEDSERVO8:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				phid->motorPositionMax = 232.992;
				phid->motorPositionMin = -22.992;
				phid->velocityMax = 1535.625;
				phid->velocityMin = 0;
				phid->accelerationMax = 4590;
				phid->accelerationMin = 18;
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		default:
			return EPHIDGET_UNEXPECTED;
	}

	//initialize triggers, set data arrays to unknown
	for (i = 0; i<phid->phid.attr.advancedservo.numMotors; i++)
	{
		phid->motorVelocityEcho[i] = PUNK_DBL;
		phid->motorPositionEcho[i] = PUNK_DBL;
		phid->motorSensedCurrent[i] = PUNK_DBL;
		//phid->motorSpeedRampingStateEcho[i] = PUNK_BOOL;
		//phid->motorEngagedStateEcho[i] = PUNK_BOOL;
	}

	//read in initial state - numMotors packets needed - try twice that in case opened elsewhere
	readtries = phid->phid.attr.advancedservo.numMotors * 2;
	while(readtries-- > 0)
	{
		CPhidget_read((CPhidgetHandle)phid);
		for (i = 0; i<phid->phid.attr.advancedservo.numMotors; i++)
			if(phid->motorVelocityEcho[i] == PUNK_DBL)
				break;
		if(i==phid->phid.attr.advancedservo.numMotors) break;
	}

	//TODO: maybe more data will be streamed from the device..
	//At this point, we can only recover (maybe) the position, set others to unknown
	for (i = 0; i<phid->phid.attr.advancedservo.numMotors; i++)
	{
		phid->motorPosition[i] = phid->motorPositionEcho[i];
		phid->motorAcceleration[i] = PUNK_DBL;
		phid->motorVelocity[i] = PUNK_DBL;
		phid->motorSpeedRampingState[i] = PUNK_BOOL;
		phid->motorEngagedState[i] = PUNK_BOOL;
	}

	return EPHIDGET_OK;
}

//dataInput - parses device packets
CPHIDGETDATA(AdvancedServo)
	int index = 0;
	double velocity = 0, position = 0, current = 0;
	double lastVelocity = 0, lastPosition = 0, lastCurrent = 0;

	if (length < 0) return EPHIDGET_INVALIDARG;
	TESTPTR(phid);
	TESTPTR(buffer);

	//Parse device packets - store data locally
	switch(phid->phid.deviceIDSpec)
	{
		case PHIDSPEC_ADVANCEDSERVO8:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				index = buffer[0];
				
				position = (unsigned short)(buffer[1] << 8) + (unsigned char)(buffer[2]);
				position = round_double(((position - 2944) / 128.0), 3);

				velocity = (signed short)(buffer[3] << 8) + (unsigned char)(buffer[4]);
				velocity = round_double(((48.0*velocity)/128.0), 2);

				current = (unsigned short)(buffer[5] << 8) + (unsigned char)(buffer[6]);
				current = round_double(((50/11) * (current/0x4000)), 4);
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		default:
			return EPHIDGET_UNEXPECTED;
	}

	//Make sure values are within defined range, and store to structure
	lastPosition = phid->motorPositionEcho[index];
	lastVelocity = phid->motorVelocityEcho[index];
	lastCurrent = phid->motorSensedCurrent[index];
	if(position > phid->motorPositionMax || position < phid->motorPositionMin)
		phid->motorPositionEcho[index] = PUNK_DBL;
	else
		phid->motorPositionEcho[index] = position;
	if(velocity > phid->velocityMax || velocity < -phid->velocityMax)
		LOG(PHIDGET_LOG_WARNING, "Phidget advanced servo recieved out of range velocity data: %lf", velocity);
	else
		phid->motorVelocityEcho[index] = velocity;
	phid->motorSensedCurrent[index] = current;

	//send out any events for changed data
	//only send a position event if the motor is engaged
	if(phid->motorPositionEcho[index] != PUNK_DBL && phid->motorPositionEcho[index] != lastPosition && phid->motorEngagedState[index] == PTRUE)
		FIRE(PositionChange, index, phid->motorPositionEcho[index]);

	if(phid->motorVelocityEcho[index] != PUNK_DBL && phid->motorVelocityEcho[index] != lastVelocity)
		FIRE(VelocityChange, index, phid->motorVelocityEcho[index]);

	if(phid->motorSensedCurrent[index] != PUNK_DBL && phid->motorSensedCurrent[index] != lastCurrent)
		FIRE(CurrentChange, index, phid->motorSensedCurrent[index]);
	
	return EPHIDGET_OK;
}

//eventsAfterOpen - sends out an event for all valid data, used during attach initialization
CPHIDGETINITEVENTS(AdvancedServo)
	int i = 0;

	for (i = 0; i<phid->phid.attr.advancedservo.numMotors; i++)
	{
		if(phid->motorSensedCurrent[i] != PUNK_DBL)
			FIRE(CurrentChange, i, phid->motorSensedCurrent[i]);
		if(phid->motorVelocityEcho[i] != PUNK_DBL)
			FIRE(VelocityChange, i, phid->motorVelocityEcho[i]);
		if(phid->motorPositionEcho[i] != PUNK_DBL && phid->motorEngagedState[i] == PTRUE)
			FIRE(PositionChange, i, phid->motorPositionEcho[i]);
	}

	return EPHIDGET_OK;
}

//getPacket - used by write thread to get the next packet to send to device
CGETPACKET_BUF(AdvancedServo)

//sendpacket - sends a packet to the device asynchronously, blocking if the 1-packet queue is full
CSENDPACKET_BUF(AdvancedServo)

//makePacket - constructs a packet using current device state
CMAKEPACKETINDEXED(AdvancedServo)
	int pwm = 0, velocity = 0, accel = 0;
	unsigned char flags = 0;

	TESTPTRS(phid, buffer);

	switch(phid->phid.deviceIDSpec)
	{
		case PHIDSPEC_ADVANCEDSERVO8:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				//have to make sure that everything to be sent has some sort of default value if the user hasn't set a value
				if(phid->motorPosition[Index] == PUNK_DBL)
					phid->motorPosition[Index] = 105; //the motors are set to this as an initial posn in the firmware - because it is a mid-range valid posn.
				if(phid->motorVelocity[Index] == PUNK_DBL)
					phid->motorVelocity[Index] = 316; //max speed of the servo Phidgets sells (0.19sec/60 degrees at 4.8V) //TODO: could bump this because it's 5V
				if(phid->motorAcceleration[Index] == PUNK_DBL)
					phid->motorAcceleration[Index] = phid->accelerationMax / 2; //mid-level acceleration
				if(phid->motorSpeedRampingState[Index] == PUNK_BOOL)
					phid->motorSpeedRampingState[Index] = PTRUE;
				if(phid->motorEngagedState[Index] == PUNK_BOOL)
					phid->motorEngagedState[Index] = PFALSE; //note motors are not engaged by default

				pwm = (int)((phid->motorPosition[Index] + 23.0) * 128.0);
				velocity = (int)((128.0*phid->motorVelocity[Index])/48.0);
				accel = (int)((128.0*phid->motorAcceleration[Index]) / (48.0*48.0));

				if(phid->motorSpeedRampingState[Index] == PFALSE)
					flags |= NO_RAMPING_FLAG;

				if(phid->motorEngagedState[Index] == PFALSE)
					flags |= STOP_FLAG;

				buffer[0] = (unsigned char)Index;
				buffer[1] = (unsigned char)((pwm >> 8) & 0xff);
				buffer[2] = (unsigned char)(pwm & 0xff);
				buffer[3] = (unsigned char)((velocity >> 8) & 0x0f);
				buffer[4] = (unsigned char)(velocity & 0xff);
				buffer[5] = (unsigned char)(accel);
				buffer[6] = flags;
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		default:
			return EPHIDGET_UNEXPECTED;
	}

	return EPHIDGET_OK;
}

// === Exported Functions === //

//create and initialize a device structure
CCREATE(AdvancedServo, PHIDDEF_ADVANCEDSERVO)

//event setup functions
CFHANDLE(AdvancedServo, PositionChange, int, double)
CFHANDLE(AdvancedServo, VelocityChange, int, double)
CFHANDLE(AdvancedServo, CurrentChange, int, double)

CGET(AdvancedServo,NumMotors,int)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	MASGN(phid.attr.advancedservo.numMotors);
}

CGETINDEX(AdvancedServo,Acceleration,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTMASGN(motorAcceleration[Index], PUNK_DBL)
	MASGN(motorAcceleration[Index])
}
CSETINDEX(AdvancedServo,Acceleration,double)
	TESTPTR(phid) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTRANGE(phid->accelerationMin, phid->accelerationMax)

	if(phid->phid.networkInfo)
		ADDNETWORKKEYINDEXED(Acceleration, "%lf");
	else
		SENDPACKETINDEXED(AdvancedServo, motorAcceleration[Index], Index);

	return EPHIDGET_OK;
}

CGETINDEX(AdvancedServo,AccelerationMax,double)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	MASGN(accelerationMax)
}

CGETINDEX(AdvancedServo,AccelerationMin,double)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	MASGN(accelerationMin)
}

CGETINDEX(AdvancedServo,MaxVelocity,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTMASGN(motorVelocity[Index], PUNK_DBL)
	MASGN(motorVelocity[Index])
}
CSETINDEX(AdvancedServo,MaxVelocity,double)
	TESTPTR(phid) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTRANGE(phid->velocityMin, phid->velocityMax)

	if(phid->phid.networkInfo)
		ADDNETWORKKEYINDEXED(MaxVelocity, "%lf");
	else
		SENDPACKETINDEXED(AdvancedServo, motorVelocity[Index], Index);

	return EPHIDGET_OK;
}

CGETINDEX(AdvancedServo,Velocity,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTMASGN(motorVelocityEcho[Index], PUNK_DBL)
	MASGN(motorVelocityEcho[Index])
}

CGETINDEX(AdvancedServo,VelocityMax,double)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	MASGN(velocityMax)
}

CGETINDEX(AdvancedServo,VelocityMin,double)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	MASGN(velocityMin)
}

CGETINDEX(AdvancedServo,MotorPosition,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTMASGN(motorPositionEcho[Index], PUNK_DBL)
	//send unknown if the motor is not engaged
	if(phid->motorEngagedState[Index] != PTRUE) {*pVal = PUNK_DBL; return EPHIDGET_UNKNOWNVAL;}
	MASGN(motorPositionEcho[Index])
}
CSETINDEX(AdvancedServo,MotorPosition,double)
	TESTPTR(phid) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTRANGE(phid->motorPositionMin, phid->motorPositionMax)

	if(phid->phid.networkInfo)
		ADDNETWORKKEYINDEXED(Position, "%lf");
	else
		SENDPACKETINDEXED(AdvancedServo, motorPosition[Index], Index);

	return EPHIDGET_OK;
}

CGETINDEX(AdvancedServo,MotorPositionMax,double)
	TESTPTRS(phid,pVal)
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	MASGN(motorPositionMax)
}

CGETINDEX(AdvancedServo,MotorPositionMin,double)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	MASGN(motorPositionMin)
}

CGETINDEX(AdvancedServo,Current,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTMASGN(motorSensedCurrent[Index], PUNK_DBL)
	MASGN(motorSensedCurrent[Index])
}

CGETINDEX(AdvancedServo,SpeedRampingOn,int)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTMASGN(motorSpeedRampingState[Index], PUNK_BOOL)
	MASGN(motorSpeedRampingState[Index]);
}
CSETINDEX(AdvancedServo,SpeedRampingOn,int)
	TESTPTR(phid) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTRANGE(PFALSE, PTRUE)

	if(phid->phid.networkInfo)
		ADDNETWORKKEYINDEXED(SpeedRampingOn, "%d");
	else
		SENDPACKETINDEXED(AdvancedServo, motorSpeedRampingState[Index], Index);

	return EPHIDGET_OK;
}

CGETINDEX(AdvancedServo,MotorOn,int)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTMASGN(motorEngagedState[Index], PUNK_BOOL)
	MASGN(motorEngagedState[Index])
}
CSETINDEX(AdvancedServo,MotorOn,int)
	TESTPTR(phid) 
	TESTDEVICETYPE(PHIDDEF_ADVANCEDSERVO)
	TESTATTACHED
	TESTINDEX(phid.attr.advancedservo.numMotors)
	TESTRANGE(PFALSE, PTRUE)

	if(phid->phid.networkInfo)
		ADDNETWORKKEYINDEXED(MotorOn, "%d");
	else
		SENDPACKETINDEXED(AdvancedServo, motorEngagedState[Index], Index);

	return EPHIDGET_OK;
}
