#include "stdafx.h"
#include "cphidgetstepper.h"
#include "cusb.h"
#include "csocket.h"
#include "cthread.h"

// === Internal Functions === //

//initAfterOpen - sets up the initial state of an object, reading in packets from the device if needed
//				  used during attach initialization - on every attach
CPHIDGETINIT(Stepper)
	int i = 0;
	int readtries = 0;

	TESTPTR(phid);
	
	//Setup max/min values
	switch(phid->phid.deviceIDSpec)
	{
		case PHIDSPEC_UNIPOLAR_STEPPER:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				phid->microSteps = 2;
				phid->motorSpeedMax = 255;
				phid->motorSpeedMin = 0;
				phid->accelerationMax = 3150;
				phid->accelerationMin = (3150 / 63);
				phid->motorPositionMax = 0x7FFFFFFFFFLL;
				phid->motorPositionMin = -0x7FFFFFFFFFLL;
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		case PHIDSPEC_BIPOLAR_STEPPER:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				phid->microSteps = 16;
				phid->motorSpeedMax = 1020;
				phid->motorSpeedMin = 0;
				phid->accelerationMax = 6300;
				phid->accelerationMin = (6300 / 63);
				phid->motorPositionMax = 0x7FFFFFFFFFLL;
				phid->motorPositionMin = -0x7FFFFFFFFFLL;
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		case PHIDSPEC_STEPPER:
		default:
			return EPHIDGET_UNEXPECTED;
	}

	//set data arrays to unknown
	for (i = 0; i<phid->phid.attr.stepper.numInputs; i++)
	{
		phid->inputState[i] = PUNK_BOOL;
	}
	for (i = 0; i<phid->phid.attr.stepper.numMotors; i++)
	{
		phid->motorSpeedEcho[i] = PUNK_DBL;
		phid->motorPositionEcho[i] = PUNK_INT64;
		phid->motorSensedCurrent[i] = PUNK_DBL;
	}

	//read in initial state - numMotors packets needed - try twice that in case opened elsewhere
	readtries = phid->phid.attr.stepper.numMotors * 2;
	while(readtries-- > 0)
	{
		CPhidget_read((CPhidgetHandle)phid);
		for (i = 0; i<phid->phid.attr.stepper.numMotors; i++)
			if(phid->motorSpeedEcho[i] == PUNK_DBL)
				break;
		if(i==phid->phid.attr.stepper.numMotors) break;
	}

	//At this point, we can only recover (maybe) the position, set others to unknown
	for (i = 0; i<phid->phid.attr.advancedservo.numMotors; i++)
	{
		phid->motorPosition[i] = phid->motorPositionEcho[i];
		phid->motorAcceleration[i] = PUNK_DBL;
		phid->motorSpeed[i] = PUNK_DBL;
		phid->motorTorque[i] = PUNK_DBL;
	}
	
	return EPHIDGET_OK;
}

//dataInput - parses device packets
CPHIDGETDATA(Stepper)
	int i = 0, j = 0;

	int index = 0;
	double speed = PUNK_DBL, current = PUNK_DBL;
	double lastSpeed = PUNK_DBL, lastCurrent = PUNK_DBL;
	long long position = PUNK_INT64, lastPosition = PUNK_INT64;
	unsigned char input[STEPPER_MAXINPUTS], lastInput[STEPPER_MAXINPUTS];

	if (length < 0) return EPHIDGET_INVALIDARG;
	TESTPTR(phid);
	TESTPTR(buffer);

	ZEROMEM(input, sizeof(input));
	ZEROMEM(lastInput, sizeof(lastInput));

	//Parse device packets - store data locally
	switch(phid->phid.deviceIDSpec)
	{
		case PHIDSPEC_UNIPOLAR_STEPPER:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				index = (buffer[0]>>6) & 0x03;

				position = ((((long long)(signed char)buffer[3]) << 32) +
					(((unsigned long long)buffer[4]) << 24) +
					(((unsigned long long)buffer[5]) << 16) +
					(((unsigned long long)buffer[6]) << 8) +
					((unsigned long long)buffer[7]));

				speed = (double)((signed short)(((buffer[0] & 0x01) ? 0xFF00 : 0x0000) | buffer[1]));
				speed = (double)round(((speed / 255.0) * phid->motorSpeedMax));

				for (i = 0, j=0x20; i<phid->phid.attr.stepper.numInputs; i++, j=j/2)
				{
					if ((buffer[0] & j))
						input[i] = PTRUE;
					else
						input[i] = PFALSE;
				}
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		case PHIDSPEC_BIPOLAR_STEPPER:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				index = (buffer[0]>>6) & 0x03;

				position = ((((long long)(signed char)buffer[3]) << 32) +
					(((unsigned long long)buffer[4]) << 24) +
					(((unsigned long long)buffer[5]) << 16) +
					(((unsigned long long)buffer[6]) << 8) +
					((unsigned long long)buffer[7]));

				speed = (double)((signed short)(((buffer[0] & 0x01) ? 0xFF00 : 0x0000) | buffer[1]));
				speed = (double)round(((speed / 255.0) * phid->motorSpeedMax));

				//value is 0-4.16 v
				current = (((unsigned char)buffer[2] * 4.16 ) / 255.0);
				current = round_double((current / (BIPOLAR_STEPPER_CURRENT_SENSE_GAIN * 0.15)), 3);
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		case PHIDSPEC_STEPPER:
		default:
			return EPHIDGET_UNEXPECTED;
	}
	
	//Make sure values are within defined range, and store to structure
	lastPosition = phid->motorPositionEcho[index];
	lastSpeed = phid->motorSpeedEcho[index];
	lastCurrent = phid->motorSensedCurrent[index];

	if(position > phid->motorPositionMax || position < phid->motorPositionMin)
		phid->motorPositionEcho[index] = PUNK_INT64;
	else
		phid->motorPositionEcho[index] = position;

	if(speed > phid->motorSpeedMax || speed < -phid->motorSpeedMax)
		LOG(PHIDGET_LOG_WARNING, "Phidget stepper recieved out of range speed data: %lf", speed);
	else
		phid->motorSpeedEcho[index] = speed;

	phid->motorSensedCurrent[index] = current;

	for (i = 0; i < phid->phid.attr.stepper.numInputs; i++)
	{
		lastInput[i] = phid->inputState[i];
		phid->inputState[i] = input[i];
	}

	//send out any events for changed data
	for (i = 0; i < phid->phid.attr.stepper.numInputs; i++)
	{
		if(phid->inputState[i] != PUNK_BOOL && phid->inputState[i] != lastInput[i])
			FIRE(InputChange, i, phid->inputState[i]);
	}

	if(phid->motorPositionEcho[index] != PUNK_INT64 && phid->motorPositionEcho[index] != lastPosition)
		FIRE(PositionChange, index, phid->motorPositionEcho[index]);

	if(phid->motorSpeedEcho[index] != PUNK_DBL && phid->motorSpeedEcho[index] != lastSpeed)
		FIRE(SpeedChange, index, phid->motorSpeedEcho[index]);

	if(phid->motorSensedCurrent[index] != PUNK_DBL && phid->motorSensedCurrent[index] != lastCurrent)
		FIRE(CurrentChange, index, phid->motorSensedCurrent[index]);

	return EPHIDGET_OK;
}

//eventsAfterOpen - sends out an event for all valid data, used during attach initialization
CPHIDGETINITEVENTS(Stepper)
	int i = 0;

	for (i = 0; i<phid->phid.attr.stepper.numInputs; i++)
	{
		if(phid->inputState[i] != PUNK_BOOL)
			FIRE(InputChange, i, phid->inputState[i]);
	}
	for (i = 0; i<phid->phid.attr.stepper.numMotors; i++)
	{
		if(phid->motorSensedCurrent[i] != PUNK_DBL)
			FIRE(CurrentChange, i, phid->motorSensedCurrent[i]);
		if(phid->motorSpeedEcho[i] != PUNK_DBL)
			FIRE(SpeedChange, i, phid->motorSpeedEcho[i]);
		if(phid->motorPositionEcho[i] != PUNK_INT64)
			FIRE(PositionChange, i, phid->motorPositionEcho[i]);
	}

	return EPHIDGET_OK;
}

//getPacket - used by write thread to get the next packet to send to device
CGETPACKET_BUF(Stepper)

//sendpacket - sends a packet to the device asynchronously, blocking if the 1-packet queue is full
CSENDPACKET_BUF(Stepper)

//makePacket - constructs a packet using current device state
CMAKEPACKETINDEXED(Stepper)
	unsigned char speed = 0, accel = 0;
	long long position = 0;

	TESTPTRS(phid, buffer);

	//have to make sure that everything to be sent has some sort of default value if the user hasn't set a value
	if(phid->motorPosition[Index] == PUNK_INT64)
		phid->motorPosition[Index] = 0;
	if(phid->motorSpeed[Index] == PUNK_DBL)
		phid->motorSpeed[Index] = phid->motorSpeedMax / 2;;
	if(phid->motorAcceleration[Index] == PUNK_DBL)
		phid->motorAcceleration[Index] = phid->accelerationMax / 2; //mid-level acceleration

	accel = (unsigned char)round((phid->motorAcceleration[Index]/phid->accelerationMax) * 63.0);
	speed = (unsigned char)round((phid->motorSpeed[Index]/phid->motorSpeedMax) * 255.0);
	position = phid->motorPosition[Index];

	switch(phid->phid.deviceIDSpec)
	{
		case PHIDSPEC_UNIPOLAR_STEPPER:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				unsigned char torque = 0;

				torque = (unsigned char)round(((phid->motorTorque[Index] * 45.0) / 100.0));

				//2-bit index
				buffer[0] = (Index << 6) & 0xc0; 
				//6-bit acceleration
				buffer[0] |= accel;
				//8-bit speed
				buffer[1] = speed;
				//1-bit for protocol expansion, 6-bit torque (bipolar)
				//range is 0-62 which is ~0-4.16 volts therefore, 0-45 =~ 0-3.0 volts
				buffer[2] = torque; 
				//40-bit position
				buffer[3] = (unsigned char)(position >> 32);
				buffer[4] = (unsigned char)(position >> 24);
				buffer[5] = (unsigned char)(position >> 16);
				buffer[6] = (unsigned char)(position >> 8);
				buffer[7] = (unsigned char)(position);
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		case PHIDSPEC_BIPOLAR_STEPPER:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				//2-bit index
				buffer[0] = (Index << 6) & 0xc0; 
				//6-bit acceleration
				buffer[0] |= accel;
				//8-bit speed
				buffer[1] = speed;
				//1-bit for protocol expansion
				buffer[2] = 0; 
				//40-bit position
				buffer[3] = (unsigned char)(position >> 32);
				buffer[4] = (unsigned char)(position >> 24);
				buffer[5] = (unsigned char)(position >> 16);
				buffer[6] = (unsigned char)(position >> 8);
				buffer[7] = (unsigned char)(position);
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		case PHIDSPEC_STEPPER:
		default:
			return EPHIDGET_UNEXPECTED;
	}

	return EPHIDGET_OK;
}

// === Exported Functions === //

//create and initialize a device structure
CCREATE(Stepper, PHIDDEF_STEPPER)

//event setup functions
CFHANDLE(Stepper, InputChange, int, int)
CFHANDLE(Stepper, PositionChange, int, long long)
CFHANDLE(Stepper, SpeedChange, int, double)
CFHANDLE(Stepper, CurrentChange, int, double)

CGET(Stepper,NumInputs,int)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	MASGN(phid.attr.stepper.numInputs)
}

CGETINDEX(Stepper,InputState,int)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numInputs)
	TESTMASGN(inputState[Index], PUNK_BOOL)
	MASGN(inputState[Index])
}

CGET(Stepper,NumMotors,int)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	MASGN(phid.attr.stepper.numMotors)
}
CGETINDEX(Stepper,Acceleration,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	TESTMASGN(motorAcceleration[Index], PUNK_DBL)
	MASGN(motorAcceleration[Index])
}
CSETINDEX(Stepper,Acceleration,double)
	TESTPTR(phid) 
	TESTINDEX(phid.attr.stepper.numMotors)
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTRANGE(phid->accelerationMin, phid->accelerationMax)

	if(phid->phid.networkInfo)
		ADDNETWORKKEYINDEXED(Acceleration, "%f");
	else
		SENDPACKETINDEXED(Stepper, motorAcceleration[Index], Index);

	return EPHIDGET_OK;
}

CGETINDEX(Stepper,AccelerationMax,double)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	MASGN(accelerationMax)
}

CGETINDEX(Stepper,AccelerationMin,double)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	MASGN(accelerationMin)
}

CGETINDEX(Stepper,MotorSpeed,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	TESTMASGN(motorSpeedEcho[Index], PUNK_DBL)
	MASGN(motorSpeedEcho[Index])
}
CSETINDEX(Stepper,MotorSpeed,double)
	TESTPTR(phid) 
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	TESTRANGE(phid->motorSpeedMin, phid->motorSpeedMax)

	if(phid->phid.networkInfo)
		ADDNETWORKKEYINDEXED(Velocity, "%lf");
	else
		SENDPACKETINDEXED(Stepper, motorSpeed[Index], Index);

	return EPHIDGET_OK;
}

CGETINDEX(Stepper,MotorSpeedMax,double)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	MASGN(motorSpeedMax)
}

CGETINDEX(Stepper,MotorSpeedMin,double)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	MASGN(motorSpeedMin)
}

CGETINDEX(Stepper,MotorPosition,long long)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	TESTMASGN(motorPositionEcho[Index], PUNK_INT64)
	MASGN(motorPositionEcho[Index])
}
CSETINDEX(Stepper,MotorPosition,long long)
	TESTPTR(phid) 
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	TESTRANGE(phid->motorPositionMin, phid->motorPositionMax)

	if(phid->phid.networkInfo)
		ADDNETWORKKEYINDEXED(Position, "%lld");
	else
		SENDPACKETINDEXED(Stepper, motorPosition[Index], Index);

	return EPHIDGET_OK;
}

CGETINDEX(Stepper,MotorPositionMax,long long)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	MASGN(motorPositionMax)
}

CGETINDEX(Stepper,MotorPositionMin,long long)
	TESTPTRS(phid,pVal) 	
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	MASGN(motorPositionMin)
}

CGETINDEX(Stepper,Current,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)

	//only the bipolar has current sense
	switch(phid->phid.deviceIDSpec)
	{
		case PHIDSPEC_BIPOLAR_STEPPER:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				TESTMASGN(motorSensedCurrent[Index], PUNK_DBL)
				MASGN(motorSensedCurrent[Index])
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		case PHIDSPEC_UNIPOLAR_STEPPER:
		case PHIDSPEC_STEPPER:
		default:
			return EPHIDGET_UNSUPPORTED;
	}
}

CSETINDEX(Stepper,Torque,double)
	TESTPTR(phid)
	TESTDEVICETYPE(PHIDDEF_STEPPER)
	TESTATTACHED
	TESTINDEX(phid.attr.stepper.numMotors)
	TESTRANGE(0, 100)

	//only the bipolar has torque
	switch(phid->phid.deviceIDSpec)
	{
		case PHIDSPEC_BIPOLAR_STEPPER:
			if ((phid->phid.deviceVersion >= 100) && (phid->phid.deviceVersion < 200))
			{
				if(phid->phid.networkInfo)
					ADDNETWORKKEYINDEXED(Torque, "%lf");
				else
					SENDPACKETINDEXED(Stepper, motorTorque[Index], Index);
			}
			else
				return EPHIDGET_UNEXPECTED;
			break;
		case PHIDSPEC_UNIPOLAR_STEPPER:
		case PHIDSPEC_STEPPER:
		default:
			return EPHIDGET_UNSUPPORTED;
	}

	return EPHIDGET_OK;
}
